/*
Copyright  2006 Altera Corporation. All rights reserved.  Altera products 
are protected under numerous U.S. and foreign patents, maskwork rights, 
copyrights and other intellectual property laws.  

This reference design file, and your use thereof, is subject to and governed
by the terms and conditions of the applicable Altera Reference Design 
License Agreement.  By using this reference design file, you indicate your
acceptance of such terms and conditions between you and Altera Corporation.
In the event that you do not agree with such terms and conditions, you may
not use the reference design file. Please promptly destroy any copies you 
have made.

This reference design file being provided on an "as-is" basis and as an 
accommodation and therefore all warranties, representations or guarantees
of any kind (whether express, implied or statutory) including, without 
limitation, warranties of merchantability, non-infringement, or fitness for
a particular purpose, are specifically disclaimed.  By making this reference
design file available, Altera expressly does not recommend, suggest or 
require that this reference design file be used in combination with any 
other product not provided by Altera
*/
  
/*******************************************************************  
* NOTICE: The information contained in this file is proprietary    *
* to Altera and is being made available to Altera's customers      *
* under strict non-disclosure agreements.  Use or disclosure of    *
* this information is permissible only under the terms of          *
* existing non-disclosure agreements.                              *
********************************************************************/  

/*----------------------------------------------------------------------------
--                                                                          --
-- Title:       channel_aligner.v                                           --
-- Rev:         Rev 1                                                       --
-- Modified:    October 17, 2005                                            --
-- Author:      David Johnson                                               --
--                                                                          --
------------------------------------------------------------------------------
--                                                                          --
-- Description: Word aligns, "byte swaps", and deskews data                 --
-- Scalable up to 32 bonded channels, modify generate statement to increase --                                                                             
-- Assert "channel_bond" to bond channels                                   --
--                                                                          -- 
------------------------------------------------------------------------------
--                                                                          --
-- Revision History:                                                        --
-- Rev 1 - Original code                                                    --
------------------------------------------------------------------------------
*/                                                                                
      
// Assert channel_bond to enable bonding
// If all pattern_detects assert within deskew_depth clocks,
// the channels are bonded and channel_aligned asserts.
// "deskew_overflow" asserts if "pattern_detects" occur in a span greater than "deskew_depth".

module	channel_aligner(
							reset_n,
							clk,
							data_in,
							channel_mask,
							channel_bond,
							pattern_detect,
							channel_aligned,
							deskew_overflow,
							data_out		
						);      
						
parameter	channel_width		= 16;
parameter	deskew_depth		= 32;   	// Make a power of 2
parameter	deskew_ptr_width	= 5;	// log2(deskew_depth)
parameter	serial_channels 	= 20;        
parameter	bonded_word_width	= channel_width * serial_channels;
parameter	deskew_depth_m1		= deskew_depth - 1;
                       

input										reset_n;
input										clk;	
input		[bonded_word_width-1:0]			data_in;
input		[serial_channels-1:0]			channel_mask;	// Ignore channel if channel_mask=1 
input										channel_bond;
input		[serial_channels-1:0]			pattern_detect;
output		reg								channel_aligned;
output		reg								deskew_overflow;
output		reg	[bonded_word_width-1:0]		data_out;

// Internal data structures:
reg			[deskew_ptr_width-1:0]			deskew_read_ptrs	[0:serial_channels-1]; 
reg			[deskew_ptr_width-1:0]			deskew_write_ptrs	[0:serial_channels-1]; 
reg			[serial_channels-1:0]			pattern_found;
wire		[bonded_word_width-1:0]			data_out_sig;	// Add pipeline to improve timing

// Register synchronous inputs
reg			[bonded_word_width-1:0]			data_in_reg;
reg			[serial_channels-1:0]			pattern_detect_reg;
reg											channel_bond_reg; 


integer	B;

always	@(posedge clk or negedge reset_n)                           
begin
	if (!reset_n)
	begin      
		data_out			<=	0;
		channel_aligned		<=	0;
		pattern_found		=	0;
		deskew_overflow		=	0; 
		data_in_reg			<=	0;
		pattern_detect_reg	<=	0;
		channel_bond_reg	<=	0;
		init_ptrs;  
	end else begin 
		data_out			<=	data_out_sig;
		data_in_reg			<=	data_in;
		pattern_detect_reg	<=	pattern_detect;
		channel_bond_reg	<=	channel_bond;
		if	(!channel_bond_reg)
		begin
			if	((&(pattern_found | channel_mask)) || (&(pattern_detect_reg | channel_mask)))
			begin
				delay_sr;
			end else begin
				init_ptrs;
			end
		end else begin
			if	((&(pattern_found | channel_mask)) || (&(pattern_detect_reg | channel_mask))) 
			begin
                delay_sr;
				if	(&(pattern_detect_reg | channel_mask)) // Special case - no deskew required
				begin
					pattern_found	= {serial_channels{1'b1}};
					channel_aligned <= 1;
				end
			end else begin	
				deskew_overflow = 0;
				for	(B=0; B<serial_channels; B=B+1)
				begin
					if	(pattern_found[B] || pattern_detect_reg[B])
					begin
						pattern_found[B]	=	1;
						deskew_write_ptrs[B]<=	deskew_write_ptrs[B] + 1;
						if	(deskew_write_ptrs[B] == deskew_depth_m1)
						begin
							deskew_overflow	= 1;
						end
					end
				end
				if	(deskew_overflow)
				begin
					pattern_found	=	0;
					init_ptrs;
				end
				channel_aligned <=	((pattern_found | channel_mask) == {serial_channels{1'b1}});
			end
		end			
	end
end
 

task	init_ptrs;
integer	A;
	begin
		for	(A=0; A<serial_channels; A=A+1)
		begin
			deskew_read_ptrs[ A]	<=	0;
			deskew_write_ptrs[A]	<=	0;
		end
	end
endtask     

task	delay_sr; 
integer	A;
	begin
		for	(A=0; A<serial_channels; A=A+1)
		begin
		    deskew_read_ptrs[ A]	<=	deskew_read_ptrs[ A] + 1;
		    deskew_write_ptrs[A]	<=	deskew_write_ptrs[A] + 1;
		end
	end
endtask   

// Instantiate each M4K or M512 section:  
// By controlling addr_out, can add insertion/deletion capability
generate
	if (serial_channels >= 1 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_1 	(.q(data_out_sig[(channel_width*1 )-1:channel_width*0 ]),.addr_in(deskew_write_ptrs[0 ]),.addr_out(deskew_read_ptrs[0 ]),.d(data_in_reg[(channel_width*1 )-1:channel_width*0 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 2 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_2 	(.q(data_out_sig[(channel_width*2 )-1:channel_width*1 ]),.addr_in(deskew_write_ptrs[1 ]),.addr_out(deskew_read_ptrs[1 ]),.d(data_in_reg[(channel_width*2 )-1:channel_width*1 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 3 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_3 	(.q(data_out_sig[(channel_width*3 )-1:channel_width*2 ]),.addr_in(deskew_write_ptrs[2 ]),.addr_out(deskew_read_ptrs[2 ]),.d(data_in_reg[(channel_width*3 )-1:channel_width*2 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 4 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_4 	(.q(data_out_sig[(channel_width*4 )-1:channel_width*3 ]),.addr_in(deskew_write_ptrs[3 ]),.addr_out(deskew_read_ptrs[3 ]),.d(data_in_reg[(channel_width*4 )-1:channel_width*3 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 5 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_5 	(.q(data_out_sig[(channel_width*5 )-1:channel_width*4 ]),.addr_in(deskew_write_ptrs[4 ]),.addr_out(deskew_read_ptrs[4 ]),.d(data_in_reg[(channel_width*5 )-1:channel_width*4 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 6 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_6 	(.q(data_out_sig[(channel_width*6 )-1:channel_width*5 ]),.addr_in(deskew_write_ptrs[5 ]),.addr_out(deskew_read_ptrs[5 ]),.d(data_in_reg[(channel_width*6 )-1:channel_width*5 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 7 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_7 	(.q(data_out_sig[(channel_width*7 )-1:channel_width*6 ]),.addr_in(deskew_write_ptrs[6 ]),.addr_out(deskew_read_ptrs[6 ]),.d(data_in_reg[(channel_width*7 )-1:channel_width*6 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 8 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_8 	(.q(data_out_sig[(channel_width*8 )-1:channel_width*7 ]),.addr_in(deskew_write_ptrs[7 ]),.addr_out(deskew_read_ptrs[7 ]),.d(data_in_reg[(channel_width*8 )-1:channel_width*7 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 9 )	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_9 	(.q(data_out_sig[(channel_width*9 )-1:channel_width*8 ]),.addr_in(deskew_write_ptrs[8 ]),.addr_out(deskew_read_ptrs[8 ]),.d(data_in_reg[(channel_width*9 )-1:channel_width*8 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 10)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_10	(.q(data_out_sig[(channel_width*10)-1:channel_width*9 ]),.addr_in(deskew_write_ptrs[9 ]),.addr_out(deskew_read_ptrs[9 ]),.d(data_in_reg[(channel_width*10)-1:channel_width*9 ]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 11)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_11	(.q(data_out_sig[(channel_width*11)-1:channel_width*10]),.addr_in(deskew_write_ptrs[10]),.addr_out(deskew_read_ptrs[10]),.d(data_in_reg[(channel_width*11)-1:channel_width*10]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 12)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_12	(.q(data_out_sig[(channel_width*12)-1:channel_width*11]),.addr_in(deskew_write_ptrs[11]),.addr_out(deskew_read_ptrs[11]),.d(data_in_reg[(channel_width*12)-1:channel_width*11]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 13)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_13	(.q(data_out_sig[(channel_width*13)-1:channel_width*12]),.addr_in(deskew_write_ptrs[12]),.addr_out(deskew_read_ptrs[12]),.d(data_in_reg[(channel_width*13)-1:channel_width*12]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 14)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_14	(.q(data_out_sig[(channel_width*14)-1:channel_width*13]),.addr_in(deskew_write_ptrs[13]),.addr_out(deskew_read_ptrs[13]),.d(data_in_reg[(channel_width*14)-1:channel_width*13]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 15)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_15	(.q(data_out_sig[(channel_width*15)-1:channel_width*14]),.addr_in(deskew_write_ptrs[14]),.addr_out(deskew_read_ptrs[14]),.d(data_in_reg[(channel_width*15)-1:channel_width*14]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 16)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_16	(.q(data_out_sig[(channel_width*16)-1:channel_width*15]),.addr_in(deskew_write_ptrs[15]),.addr_out(deskew_read_ptrs[15]),.d(data_in_reg[(channel_width*16)-1:channel_width*15]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 17)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_17	(.q(data_out_sig[(channel_width*17)-1:channel_width*16]),.addr_in(deskew_write_ptrs[16]),.addr_out(deskew_read_ptrs[16]),.d(data_in_reg[(channel_width*17)-1:channel_width*16]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 18)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_18	(.q(data_out_sig[(channel_width*18)-1:channel_width*17]),.addr_in(deskew_write_ptrs[17]),.addr_out(deskew_read_ptrs[17]),.d(data_in_reg[(channel_width*18)-1:channel_width*17]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 19)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_19	(.q(data_out_sig[(channel_width*19)-1:channel_width*18]),.addr_in(deskew_write_ptrs[18]),.addr_out(deskew_read_ptrs[18]),.d(data_in_reg[(channel_width*19)-1:channel_width*18]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 20)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_20	(.q(data_out_sig[(channel_width*20)-1:channel_width*19]),.addr_in(deskew_write_ptrs[19]),.addr_out(deskew_read_ptrs[19]),.d(data_in_reg[(channel_width*20)-1:channel_width*19]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 21)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_21	(.q(data_out_sig[(channel_width*21)-1:channel_width*20]),.addr_in(deskew_write_ptrs[20]),.addr_out(deskew_read_ptrs[20]),.d(data_in_reg[(channel_width*21)-1:channel_width*20]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 22)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_22	(.q(data_out_sig[(channel_width*22)-1:channel_width*21]),.addr_in(deskew_write_ptrs[21]),.addr_out(deskew_read_ptrs[21]),.d(data_in_reg[(channel_width*22)-1:channel_width*21]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 23)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_23	(.q(data_out_sig[(channel_width*23)-1:channel_width*22]),.addr_in(deskew_write_ptrs[22]),.addr_out(deskew_read_ptrs[22]),.d(data_in_reg[(channel_width*23)-1:channel_width*22]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 24)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_24	(.q(data_out_sig[(channel_width*24)-1:channel_width*23]),.addr_in(deskew_write_ptrs[23]),.addr_out(deskew_read_ptrs[23]),.d(data_in_reg[(channel_width*24)-1:channel_width*23]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 25)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_25	(.q(data_out_sig[(channel_width*25)-1:channel_width*24]),.addr_in(deskew_write_ptrs[24]),.addr_out(deskew_read_ptrs[24]),.d(data_in_reg[(channel_width*25)-1:channel_width*24]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 26)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_26	(.q(data_out_sig[(channel_width*26)-1:channel_width*25]),.addr_in(deskew_write_ptrs[25]),.addr_out(deskew_read_ptrs[25]),.d(data_in_reg[(channel_width*26)-1:channel_width*25]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 27)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_27	(.q(data_out_sig[(channel_width*27)-1:channel_width*26]),.addr_in(deskew_write_ptrs[26]),.addr_out(deskew_read_ptrs[26]),.d(data_in_reg[(channel_width*27)-1:channel_width*26]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 28)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_28	(.q(data_out_sig[(channel_width*28)-1:channel_width*27]),.addr_in(deskew_write_ptrs[27]),.addr_out(deskew_read_ptrs[27]),.d(data_in_reg[(channel_width*28)-1:channel_width*27]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 29)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_29	(.q(data_out_sig[(channel_width*29)-1:channel_width*28]),.addr_in(deskew_write_ptrs[28]),.addr_out(deskew_read_ptrs[28]),.d(data_in_reg[(channel_width*29)-1:channel_width*28]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 30)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_30	(.q(data_out_sig[(channel_width*30)-1:channel_width*29]),.addr_in(deskew_write_ptrs[29]),.addr_out(deskew_read_ptrs[29]),.d(data_in_reg[(channel_width*30)-1:channel_width*29]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 31)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_31	(.q(data_out_sig[(channel_width*31)-1:channel_width*30]),.addr_in(deskew_write_ptrs[30]),.addr_out(deskew_read_ptrs[30]),.d(data_in_reg[(channel_width*31)-1:channel_width*30]),.we(1),.clk1(clk),.clk2(clk));
	if (serial_channels >= 32)	ram_dual	#(channel_width,deskew_depth,deskew_ptr_width)	ram_dual_32	(.q(data_out_sig[(channel_width*32)-1:channel_width*31]),.addr_in(deskew_write_ptrs[31]),.addr_out(deskew_read_ptrs[31]),.d(data_in_reg[(channel_width*32)-1:channel_width*31]),.we(1),.clk1(clk),.clk2(clk));

	
endgenerate




endmodule       
