//--------------------------------------------------------------------------//
//                                                                          //
// Title:       skew_adjust_bts.v     	        		                	//
// Rev:         Rev 1                                                       //
// Created:     March 8, 2006		                                       	//
// Author:		David Johnson												//
//                                                                          //
//--------------------------------------------------------------------------//
//                                                                          //
// Description: 	Introduces skew into the serial datastream by modifying	//
//              	the parallel data stream.  Resolution is 1 bit	period	//
//                                                                          //
//--------------------------------------------------------------------------//
//                                                                          //
// Revision History:                                                        //
//                                                                          //
// Rev 1:       First cut													//
//              					                        				//
//                                                                          //
//----------------------------------------------------------------------------
//------ 1 ------- 2 ------- 3 ------- 4 ------- 5 ------- 6 ------- 7 ------7
//------ 0 ------- 0 ------- 0 ------- 0 ------- 0 ------- 0 ------- 0 ------8
//----------------------------------------------------------------------------
//Copyright  2006 Altera Corporation. All rights reserved.  Altera products
//are protected under numerous U.S. and foreign patents, maskwork rights,
//copyrights and other intellectual property laws.
//                 
//This reference design file, and your use thereof, is subject to and
//governed by the terms and conditions of the applicable Altera Reference
//Design License Agreement.  By using this reference design file, you
//indicate your acceptance of such terms and conditions between you and
//Altera Corporation.  In the event that you do not agree with such terms and
//conditions, you may not use the reference design file. Please promptly                         
//destroy any copies you have made.
//
//This reference design file being provided on an "as-is" basis and as an
//accommodation and therefore all warranties, representations or guarantees
//of any kind (whether express, implied or statutory) including, without
//limitation, warranties of merchantability, non-infringement, or fitness for
//a particular purpose, are specifically disclaimed.  By making this
//reference design file available, Altera expressly does not recommend,
//suggest or require that this reference design file be used in combination 
//with any other product not provided by Altera.
                                                                      
// Module will add up to num_barrels * channel_width bits of skew delay.
// Resolution is to 1 serial bit period for non 8B/10B data.
// If sending control characters and data (8B/10B), pcs expects aligned data, 
// therefore only provide skew numbers which are multiples of 8.
// Send multiple parallel data channels concatenated on "datain".
// Send multiple delay skew values concatenated on "skew".
// Multiple data channels are output on "dataout", for example for direct connection
// to the tx_datain on the alt2gxb XCVRs

module	skew_adjust_bts	(
							reset_n,
							clk,
							datain,
							dataout,
							skew    
						); 
						
parameter	serial_channels		=	1;    
parameter	channel_width		=	8;
parameter	channel_skew_width	=	3;				// >= ceil(log2(channel_width*(num_barrels-1))) 
parameter	num_barrels			=	2;				// set >=2 (Number of barrels in barrel shifter per channel)
parameter	data_width			=	serial_channels * channel_width;
parameter	skew_width			=	serial_channels * channel_skew_width;                                                                        
parameter	skew_msb_1st		=	0;				// skew_msb_1st=1 (eg Source Synchronous LVDS or SONET are MSB 1st)
													// skew_msb_1st=0 (eg 8B/10B XCVRS with LSB 1st) 

input									reset_n;
input									clk;
input	[data_width-1:0]				datain;
output	[data_width-1:0]				dataout;
input	[skew_width-1:0]				skew;

wire									reset_n;
wire									clk;
wire	[data_width-1:0]				datain;
reg		[data_width-1:0]				dataout;
wire	[skew_width-1:0]				skew;

// Internal data structures:
// barrel_reg can also be implemented using M4K or M512
reg		[(num_barrels*data_width)-1:0]	barrel_reg;
integer									A,B,C;   
reg		[channel_skew_width-1:0]		skew_w;


	always	@(posedge clk or negedge reset_n)
	begin
		if	(!reset_n)     
		begin
			barrel_reg	<=	0;
			dataout		<=	0;
		end else begin
			for (A=0; A<serial_channels; A=A+1) // Scalable multi-stage barrel register:
			begin
				for (B=0; B<num_barrels; B=B+1)
				begin
					for (C=0; C<channel_width; C=C+1)
					begin
						if	(skew_msb_1st)
						begin
							if (B==0)
							begin
								barrel_reg[(A*channel_width*num_barrels)+(B*channel_width)+C] <= datain[(A*channel_width)+C];
							end else begin
								barrel_reg[(A*channel_width*num_barrels)+(B*channel_width)+C]  <= barrel_reg[(A*channel_width*num_barrels)+((B-1)*channel_width)+C];
							end								
						end else begin
							if (B==num_barrels-1)
							begin
								barrel_reg[(A*channel_width*num_barrels)+(B*channel_width)+C] <= datain[(A*channel_width)+C];
							end else begin
								barrel_reg[(A*channel_width*num_barrels)+(B*channel_width)+C]  <= barrel_reg[(A*channel_width*num_barrels)+((B+1)*channel_width)+C];
							end
						end
					end
				end
			end
			for	(A=0; A<serial_channels; A=A+1) // Add skew delay:
			begin   
				// Get skew for channel A
				for	(B=0; B<channel_skew_width; B=B+1)
				begin				
					skew_w[B] = skew[(A*channel_skew_width)+B];				// Get skew for channel "A"
				end  
				// Apply skew delay to channel A
				for	(B=0; B<channel_width; B=B+1)							// Add skew delay for channel "A"
				begin	
					dataout[(A*channel_width)+B]	<=	barrel_reg[(A*channel_width*num_barrels)+B+skew_w];
				end		
			end
		end                                                      
	end	        
	
                                                              

endmodule
