/*
Copyright  2009 Altera Corporation. All rights reserved.  Altera products 
are protected under numerous U.S. and foreign patents, maskwork rights, 
copyrights and other intellectual property laws.  

This reference design file, and your use thereof, is subject to and governed
by the terms and conditions of the applicable Altera Reference Design 
License Agreement.  By using this reference design file, you indicate your
acceptance of such terms and conditions between you and Altera Corporation.
In the event that you do not agree with such terms and conditions, you may
not use the reference design file. Please promptly destroy any copies you 
have made.

This reference design file being provided on an "as-is" basis and as an 
accommodation and therefore all warranties, representations or guarantees
of any kind (whether express, implied or statutory) including, without 
limitation, warranties of merchantability, non-infringement, or fitness for
a particular purpose, are specifically disclaimed.  By making this reference
design file available, Altera expressly does not recommend, suggest or 
require that this reference design file be used in combination with any 
other product not provided by Altera
*/
  
/*******************************************************************  
* NOTICE: The information contained in this file is proprietary    *
* to Altera and is being made available to Altera's customers      *
* under strict non-disclosure agreements.  Use or disclosure of    *
* this information is permissible only under the terms of          *
* existing non-disclosure agreements.                              *
********************************************************************/  

/*----------------------------------------------------------------------------
--                                                                          --
-- Title:       word_aligner.v                                              --
-- Rev:         Rev 3                                                       --
-- Modified:    September 16, 2009                                               --
-- Author:      Altera                                                      --
--                                                                          --
------------------------------------------------------------------------------
--                                                                          --
-- Description: Word aligns and "byte swaps"                                --
--                                                                          --                                                                             
-- Assert "en_cdet" to word align                                           -- 
-- Performs "byte swapping" during word alignment                           --
--                                                                          -- 
------------------------------------------------------------------------------
--                                                                          --
-- Revision History:                                                        --
-- Rev 1 - Original code                                                    --
-- Rev 2 - Replaced pattern match decision with one-hot                     --
--         No mimic patterns should occur while word aligning               --
--         Brent Fairbanks analyzed the code and added Gregg Baeckler's     --
--         "one-hot" scalable decoder to minimize area and improve timing   --
-- Rev 3 - Added back legacy code for mux_sel_step other than 1             --
--         to support byte order function                                   --
------------------------------------------------------------------------------
*/                                                                                
      
	
// Input big endian non word aligned data, lsb transmitted 1st.
// Performs word alignment and byte reorder based on sync_pattern
// This module mixes blocking & non-blocking to
// simplify scaling the architecture.

// On detection of sync_pattern anywhere in the serial stream, resync will assert.
// pattern_detect will assert every time sync_pattern is on the current word boundary,
// and normally won't assert during re-alignment (resync will assert)
// If aligned data isn't required immediately upon detection,
// cut the timing path from mux_sel to data_reg2 to improve static timing analysis slack.


module onehot_to_bin (onehot, bin);

parameter ONEHOT_WIDTH = 16;
parameter BIN_WIDTH = log2(ONEHOT_WIDTH-1);

// baeckler - 12-12-2006
// helper function to compute LOG base 2
//
// NOTE - This is a somewhat abusive definition of LOG2(v) as the
//   number of bits required to represent "v".  So log2(256) will be
//   9 rather than 8 (256 = 9'b1_0000_0000).  I apologize for any
//   confusion this may cause.
//

function integer log2;
  input integer val;
  begin
	 log2 = 0;
	 while (val > 0) begin
	    val = val >> 1;
		log2 = log2 + 1;
	 end
  end
endfunction

input [ONEHOT_WIDTH-1:0] onehot;
output [BIN_WIDTH-1:0] bin;

genvar i,j;
generate
	for (j=0; j<BIN_WIDTH; j=j+1)
	begin : jl
		wire [ONEHOT_WIDTH-1:0] tmp_mask;
		for (i=0; i<ONEHOT_WIDTH; i=i+1)
		begin : il
			assign tmp_mask[i] = i[j];
		end	
		assign bin[j] = |(tmp_mask & onehot);
	end
endgenerate

endmodule


module	word_aligner	(
							reset_n,
							clk,
							en_cdet,
							data_in,  
							sync_pattern,
							pattern_detect,
							resync,			// Assert when a new pattern alignment is detected
							data_out		
						);      

parameter	channel_width		= 16;
parameter	sync_pattern_width	= 16;	// F628
parameter	mux_sel_width		= 4;	// log2(channel_width) (round any fraction up)    
parameter	mux_sel_step		= 1;	// If only performing byte reorder function, set to deserializer_width (channel_width / num_endians)
parameter	level_mode			= 1; 	// 1=level mode, 0=ec_cdet is edge mode
parameter	channel_width_2x	= channel_width * 2;   

input									reset_n;
input									clk;	
input									en_cdet;	
input		[channel_width-1:0]			data_in;
input		[sync_pattern_width-1:0]	sync_pattern;
output		reg							pattern_detect;
output		reg							resync;
output		reg	[channel_width-1:0]		data_out;

// Internal data structures:    	
reg			[channel_width-1:0]			data_reg2;
reg			[channel_width_2x-1:0]		barrel_reg;
reg			[mux_sel_width-1:0]			mux_sel;		
reg										en_cdet_prev;
reg										resync_w;
reg										en_cdet_reg;
reg			[channel_width-1:0]			match;
reg										en_cdet_edge;

integer	A;
	always @(posedge clk or negedge reset_n)
	begin    
		if	(!reset_n)
		begin 
			barrel_reg		<=	0;  
			data_reg2		<=	0;
			data_out		<=	0; 
			mux_sel			<=	0;
			pattern_detect	<=	0;
			en_cdet_reg		<=	0;
			en_cdet_prev	<=	0;
			match			<=	0;
			resync			<=	0;
			resync_w		=	0;
			en_cdet_edge	<=	0;
		end else begin
			if	(en_cdet_reg && (!en_cdet_prev))
			begin
				en_cdet_edge	<=	1;
			end	else if (pattern_detect)
			begin
				en_cdet_edge	<=	0;
			end	
			barrel_reg		<=	{barrel_reg[channel_width-1:0],data_in};
			if (mux_sel_step == 1)
			begin
				sync_detect;
			end else begin
				sync_detect_mss;
			end
			word_align;
			data_out		<=	data_reg2;
			pattern_detect	<=	(data_reg2[sync_pattern_width-1:0] == sync_pattern[sync_pattern_width-1:0]);
			en_cdet_reg		<=	en_cdet;		// Pipeline to partition design better
			en_cdet_prev	<=	en_cdet_reg;
		end
	end

	wire [mux_sel_width-1:0]			mux_sel_z;		
	onehot_to_bin oh (.onehot(match), .bin(mux_sel_z));
	defparam oh .ONEHOT_WIDTH = channel_width;

	task	sync_detect;
	integer	A,B;
	reg		[channel_width-1:0]	match_w;	// blocking
		begin 
			// Word align pattern detector:
			match_w	= ~0;
			for (A=0; A<=channel_width-1; A=A+mux_sel_step)
			begin
				for (B=0; B<=sync_pattern_width-1; B=B+1)
				begin
					if (barrel_reg[A+B] != sync_pattern[B])
					begin
						match_w[A] = 0; 
					end               
				end
			end
			
			match	<=	match_w;		 
			resync_w = 0;
			
			if (match != 0)
			begin
				if  ((en_cdet_reg && level_mode) || (en_cdet_edge && (!level_mode)))
				begin
					mux_sel <= mux_sel_z;
				end
				resync_w = 1;
			end

			resync		<= resync_w;
		end
	endtask
	
// Use if mux_sel_step isn't equal to 1
// This will perform word alignment and byte order functionality
	task	sync_detect_mss;
	integer	A,B;
	reg		[channel_width-1:0]	match_w;	// blocking
		begin 
			// Word align pattern detector:
			match_w	= ~0;
			for (A=0; A<=channel_width-1; A=A+mux_sel_step)
			begin
				for (B=0; B<=sync_pattern_width-1; B=B+1)
				begin
					if (barrel_reg[A+B] != sync_pattern[B])
					begin
						match_w[A] = 0; 
					end               
				end
			end
			match		<=	match_w;		 
			resync_w	= 0;
			for (A=0; A<=channel_width-1; A=A+mux_sel_step)
			begin
				begin
					if	(match[A])
					begin
						if	((en_cdet_reg && level_mode) || (en_cdet_edge && (!level_mode)))
						begin
							mux_sel		<= A;
						end
						resync_w		= 1;
					end
				end
			end
			resync		<= resync_w;
		end
	endtask	
	     
	task	word_align;
	integer A;
		begin
			for (A=0; A<=channel_width-1; A=A+1)
			begin
				data_reg2[A]	<=	barrel_reg[A+mux_sel];	
			end
		end
	endtask



endmodule       
